"""Historical data objects"""
import collections


def namedtuple(typename, field_names):
    """
    Creates a namedtuple object that silently ignores unknown keys.

    This ensures compatibility with changing API.
    """
    base = collections.namedtuple("Base", field_names)
    return type(
        typename,
        (base,),
        {
            "__slots__": (),
            "__new__": lambda cls, *args, **kwargs: base.__new__(
                cls, *args, **{k: v for k, v in kwargs.items() if k in base._fields}
            ),
        },
    )


HDBlock = namedtuple(
    "HDBlock",
    [
        "number",
        "hash",
        "parentHash",
        "nonce",
        "sha3Uncles",
        "logsBloom",
        "transactionsRoot",
        "stateRoot",
        "receiptsRoot",
        "miner",
        "difficulty",
        "totalDifficulty",
        "extraData",
        "size",
        "gasLimit",
        "gasUsed",
        "transactions",
        "timestamp",
        "uncles",
    ],
)

HDTransaction = namedtuple(
    "HDTransaction",
    [
        "blockHash",
        "blockNumber",
        # replace "from" with "sender" to suppress Python error on keyword
        "sender",
        "gasLimit",
        "hash",
        "nonce",
        "transactionIndex",
        "to",
        "value",
        "gasPrice",
        "input",
        "confirmations",
        "error",
    ],
)

# if native asset balance,contract is set to address of the contract
HDBalance = namedtuple(
    "HDBalance",
    [
        "id",
        "address",
        "contract",
        "balance",
        "transactionHash",
        "transactionIndex",
        "blockHash",
        "blockNumber",
        "timestamp",
        "isMainChain",
    ],
)

# result is a list of HDTransaction objects
HDTransactions = namedtuple("HDTransactions", ["result", "next_cursor"])

HDStatus = namedtuple("HDStatus", ["lowest", "highest", "latest"])



from upvest.authentication import OAuth
from upvest.config import UPVEST_API_TARGET
from upvest.model import Assets, Wallets
from upvest.utils import verify_echo


class UpvestClienteleAPI:
    def __init__(self, client_id, client_secret, username, password, base_url=None, user_agent=None):
        base_url = base_url or UPVEST_API_TARGET
        self.auth_instance = OAuth(
            client_id=client_id,
            client_secret=client_secret,
            username=username,
            password=password,
            base_url=base_url,
            user_agent=user_agent,
        )
        self.assets = Assets(self.auth_instance)
        self.wallets = Wallets(self.auth_instance)

    def check_auth(self):
        verify_echo(self.auth_instance, "/clientele/echo-oauth2")



import hashlib
import hmac
import time
from urllib.parse import urljoin
from upvest.__pkginfo__ import DEFAULT_USERAGENT

import requests

from upvest.config import API_VERSION, ENCODING, GRANT_TYPE, OAUTH_PATH, SCOPE
from upvest.exceptions import InvalidRequest


class KeyAuth:
    def __init__(self, api_key=None, api_secret=None, api_passphrase=None, base_url=None, user_agent=None):
        self.api_key = api_key
        self.api_secret = api_secret
        self.api_passphrase = api_passphrase
        self.base_url = base_url
        self.user_agent = user_agent or DEFAULT_USERAGENT

    def get_headers(self, method, path, body=None):
        timestamp = str(int(time.time()))
        # Compose the message as a concatenation of all info we are sending along with the request
        message = timestamp + method + API_VERSION + path + (body or "")
        # Generate signature, in order to prevent manipulation of payload in flight
        signature = hmac.new(
            str.encode(self.api_secret), msg=message.encode(ENCODING), digestmod=hashlib.sha512
        ).hexdigest()
        # Generate message headers
        headers = {
            "Content-Type": "application/json",
            "X-UP-API-Key": self.api_key,
            "X-UP-API-Signature": signature,
            "X-UP-API-Timestamp": timestamp,
            "X-UP-API-Passphrase": self.api_passphrase,
            "X-UP-API-Signed-Path": API_VERSION + path,
        }
        return headers


class OAuth:
    def __init__(
        self, client_id=None, client_secret=None, username=None, password=None, base_url=None, user_agent=None
    ):
        self.path = API_VERSION + OAUTH_PATH
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.base_url = base_url
        self.user_agent = user_agent or DEFAULT_USERAGENT

        self.access_token = None
        self.access_token_expiry = None

    def get_token(self):
        # allow for some clock drift
        if self.access_token and time.time() < self.access_token_expiry - 3600:
            return self.access_token

        # Set header content-type to x-www-form-urlencoded
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body = {
            "grant_type": GRANT_TYPE,
            "scope": SCOPE,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": self.username,
            "password": self.password,
        }

        # send x-www-form-urlencoded payload to clientele API.
        response = requests.post(urljoin(self.base_url, self.path), data=body, headers=headers)
        if response.status_code >= 300:
            raise InvalidRequest(response)

        # Retrieve and return OAuth token
        resp_json = response.json()
        self.access_token = resp_json["access_token"]
        self.access_token_expiry = time.time() + resp_json["expires_in"]
        return self.access_token

    def get_headers(self, **_):
        return {"Authorization": f"Bearer {self.get_token()}", "Content-Type": "application/json"}



import base64

from upvest.authentication import KeyAuth
from upvest.config import UPVEST_API_TARGET
from upvest.exceptions import RecoveryFailedError
from upvest.model import Assets, Users, Webhooks, HistoricalData
from upvest.utils import Request, Response, verify_echo


class UpvestTenancyAPI:
    def __init__(self, api_key, api_secret, api_passphrase, base_url=None, user_agent=None):
        base_url = base_url or UPVEST_API_TARGET
        self.auth_instance = KeyAuth(
            api_key=api_key,
            api_secret=api_secret,
            api_passphrase=api_passphrase,
            base_url=base_url,
            user_agent=user_agent,
        )
        self.users = Users(self.auth_instance)
        self.assets = Assets(self.auth_instance)
        self.webhooks = Webhooks(self.auth_instance)
        self.historical = HistoricalData(self.auth_instance)

    def check_auth(self):
        verify_echo(self.auth_instance, "/tenancy/echo-signed")

    def recover_with_seed(self, seed, seedhash, user_id, password):
        path = "/tenancy/recover/"

        # Set username and password for the user
        body = {"seed": seed, "seedhash": seedhash, "user_id": user_id, "password": password}
        response = Response(Request().post(auth_instance=self.auth_instance, path=path, body=body))
        if not response.data["success"]:
            raise RecoveryFailedError("Recovery unsuccessful")

    def recover(self, recovery_kit_base64, private_key_base64, new_password):
        try:
            from google.protobuf.json_format import MessageToDict
            from nacl.public import PrivateKey, SealedBox
            from upvest.proto import RecoveryKit
        except (ImportError, ModuleNotFoundError):
            raise Exception("Please install via pip install upvest[recovery]")

        cipher = base64.b64decode(recovery_kit_base64)

        pk = PrivateKey(base64.b64decode(private_key_base64))
        proto = SealedBox(pk).decrypt(cipher)

        kit = RecoveryKit()
        kit.ParseFromString(proto)

        seed = MessageToDict(kit.seed, including_default_value_fields=True)

        self.recover_with_seed(seed=seed, seedhash=kit.seedhash, user_id=kit.user_id, password=new_password)



import json
from urllib.parse import urljoin

import requests

from upvest.config import API_VERSION
from upvest.exceptions import InvalidRequest, AuthenticationError


class Response:
    def __init__(self, result):
        self.status_code = result.status_code
        self.raw = result
        self.data = None
        if result.content:
            self.json = result.json()
            self.data = self.json.get("results", self.json)


# Request object
class Request:
    def __init__(self):
        pass

    def _check(self, value):
        if isinstance(value, int):
            pass
        elif isinstance(value, str):
            try:
                value.encode("ascii")
            except UnicodeEncodeError:
                raise ValueError("Forbidden characters present, please remove")
        elif isinstance(value, dict):
            for val in value.values():
                self._check(val)
        elif isinstance(value, list):
            for val in value:
                self._check(val)
        else:
            raise ValueError("no valid JSON structure given")

    def _request(self, auth_instance, method, path, body=None, params=None):
        if body is not None:
            self._check(body)
            body = json.dumps(body)
        # Instantiate the respectively needed auth instance
        authenticated_headers = auth_instance.get_headers(method=method, path=path, body=body)
        authenticated_headers["User-Agent"] = auth_instance.user_agent
        # Execute request with authenticated headers
        request_url = urljoin(auth_instance.base_url, API_VERSION + path)
        response = requests.request(method, request_url, data=body, params=params, headers=authenticated_headers)
        if response.status_code >= 300:
            raise InvalidRequest(response)
        else:
            return response

    def post(self, **req_params):
        req_params["method"] = "POST"
        return self._request(**req_params)

    def get(self, **req_params):
        req_params["method"] = "GET"
        return self._request(**req_params)

    def patch(self, **req_params):
        req_params["method"] = "PATCH"
        return self._request(**req_params)

    def delete(self, **req_params):
        req_params["method"] = "DELETE"
        return self._request(**req_params)


def verify_echo(auth_instance, path):
    body = {"echo": "test"}
    try:
        resp = Response(Request().post(auth_instance=auth_instance, path=path, body=body))
    except InvalidRequest as e:
        if hasattr(e.args[0], "status_code") and e.args[0].status_code in (401, 403):
            raise AuthenticationError("Unauthorized/Forbidden")

        if isinstance(e.args[0], str) and json.loads(e.args[0])["error"] == "invalid_grant":
            raise AuthenticationError("Invalid grant")

        # Other, unexpected error
        raise
    if resp.data["echo"] != "test":
        raise AuthenticationError("Unexpected echo response")



import http
import hashlib
from base64 import b64encode
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from upvest.config import API_VERSION
from upvest.utils import Request, Response
from upvest.historical import HDTransaction, HDTransactions, HDBlock, HDBalance, HDStatus


# Endpoint related objects
class UserInstance:
    def __init__(self, auth_instance, username, wallets=None, recovery_kit=None):
        self.path = "/tenancy/users/"
        self.auth_instance = auth_instance
        self.username = username
        self.recovery_kit = recovery_kit
        if wallets is None:
            self.wallets = []
        else:
            self.wallets = [WalletInstance(auth_instance, **w) for w in wallets]

    def update(self, password, new_password):
        # Provide current and new password
        body = {"old_password": password, "new_password": new_password}
        response = Response(
            Request().patch(auth_instance=self.auth_instance, path=self.path + self.username, body=body)
        )
        username = response.data["username"]
        return UserInstance(response, username)

    def delete(self):
        response = Response(
            Request().delete(auth_instance=self.auth_instance, path=self.path + self.username, body=None)
        )
        return response.status_code == 204

    def __eq__(self, other):
        return self.username == other.username


class Users:
    def __init__(self, auth_instance):
        self.path = "/tenancy/users/"
        self.auth_instance = auth_instance

    def create(self, username, password, raw_recovery_kit=False, asset_ids=None):
        # Set username and password for the user
        body = {"username": username, "password": password, "raw": raw_recovery_kit, "asset_ids": asset_ids or []}
        response = Response(Request().post(auth_instance=self.auth_instance, path=self.path, body=body))
        username = response.data["username"]
        recovery_kit = response.data["recoverykit"]
        return UserInstance(self.auth_instance, username, recovery_kit=recovery_kit)

    def get(self, username):
        response = Response(Request().get(auth_instance=self.auth_instance, path=self.path + username))
        username = response.data["username"]
        return UserInstance(self.auth_instance, username, wallets=response.data["wallets"])

    def list(self, count):
        # Retrieve subset of users
        MAX_PAGE_SIZE = 100
        array_of_users = []
        path = f"/tenancy/users/?page_size={MAX_PAGE_SIZE}"
        listed_count = 0

        while listed_count < count:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for user in response.data:
                array_of_users.append(UserInstance(self.auth_instance, user["username"], wallets=user["wallets"]))
                listed_count += 1
                if listed_count >= count:
                    break
            if response.raw.json()["next"]:
                page_size = min(MAX_PAGE_SIZE, count - listed_count)
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [page_size]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_users

    def all(self):
        MAX_PAGE_SIZE = 100
        array_of_users = []
        path = "/tenancy/users/?{MAX_PAGE_SIZE}"

        while True:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for user in response.data:
                array_of_users.append(UserInstance(self.auth_instance, user["username"], wallets=user["wallets"]))
            if response.raw.json()["next"]:
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [MAX_PAGE_SIZE]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_users


class AssetInstance:
    def __init__(self, **asset_attr):
        self.id = asset_attr["id"]
        self.name = asset_attr["name"]
        self.symbol = asset_attr["symbol"]
        self.exponent = asset_attr["exponent"]
        self.protocol = asset_attr["protocol"]
        self.metadata = asset_attr["metadata"]


class Assets:
    def __init__(self, auth_instance):
        self.path = "/assets/"
        self.auth_instance = auth_instance

    def all(self):
        MAX_PAGE_SIZE = 100
        array_of_assets = []
        path = "/assets/?{MAX_PAGE_SIZE}"

        while True:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for asset in response.data:
                array_of_assets.append(AssetInstance(**asset))
            if response.raw.json()["next"]:
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [MAX_PAGE_SIZE]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_assets


class ECDSASignature:
    def __init__(self, signed_hash, j):
        self.algorithm = j["algorithm"]
        if self.algorithm != "ECDSA":
            raise ValueError(f"unsupported algorithm: {self.algorithm}")

        self.curve = j["curve"]
        self.x = j["public_key"]["x"]
        self.y = j["public_key"]["y"]
        self.r = j["r"]
        self.s = j["s"]
        self.signed_hash = signed_hash

        if self.curve == "secp256k1":
            self.recover = j["recover"]
        else:
            # TODO: is pubkey recovery possible for other curves?
            pass


class WalletInstance:
    def __init__(self, auth_instance, **wallet_attr):
        self.auth_instance = auth_instance
        self.transactions = Transactions(auth_instance, wallet_attr["id"])
        self.id = wallet_attr["id"]
        self.path = f"/kms/wallets/{self.id}"
        self.balances = wallet_attr["balances"]
        self.protocol = wallet_attr["protocol"]
        self.address = wallet_attr["address"]
        self.status = wallet_attr["status"]

    def sign(self, password, message=None, to_sign_hash=None, hash_algorithm="SHA256"):
        if to_sign_hash is None and message is not None:
            if not isinstance(message, (bytes, bytearray)):
                raise TypeError("message argument is not a bytes-like object")

            if hash_algorithm == "SHA256":
                to_sign_hash = hashlib.sha256(message).digest()
            else:
                raise ValueError(f"unsupported hash_algorithm: {hash_algorithm}")

        if to_sign_hash is None:
            raise ValueError(f"neither message nor to_sign_hash were provided")

        body = {"password": password, "to_sign": str(b64encode(to_sign_hash), "UTF-8"), "output_format": "int"}
        rsp = Response(Request().post(auth_instance=self.auth_instance, path=f"{self.path}/sign", body=body))
        return ECDSASignature(signed_hash=to_sign_hash, j=rsp.data)


class Wallets:
    def __init__(self, auth_instance):
        self.path = "/kms/wallets/"
        self.auth_instance = auth_instance

    def create(self, asset_id, password):
        # Get desired asset id from assets list
        # Provide password and asset_id for wallet creation
        body = {"password": password, "asset_id": asset_id}
        response = Response(Request().post(auth_instance=self.auth_instance, path=self.path, body=body))
        return WalletInstance(self.auth_instance, **response.data)

    def get(self, wallet_id):
        # Retrieve specific wallet for a user
        response = Response(Request().get(auth_instance=self.auth_instance, path=self.path + wallet_id))
        return WalletInstance(self.auth_instance, **response.data)

    def list(self, count):
        # Retrieve subset of wallets
        MAX_PAGE_SIZE = 100
        array_of_wallets = []
        path = f"/kms/wallets/?page_size={MAX_PAGE_SIZE}"
        listed_count = 0

        while listed_count < count:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for wallet in response.data:
                array_of_wallets.append(WalletInstance(self.auth_instance, **wallet))
                listed_count += 1
                if listed_count >= count:
                    break
            if response.raw.json()["next"]:
                page_size = min(MAX_PAGE_SIZE, count - listed_count)
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [page_size]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_wallets

    def all(self):
        # Retrieve subset of wallets
        MAX_PAGE_SIZE = 100
        array_of_wallets = []
        path = f"/kms/wallets/?page_size={MAX_PAGE_SIZE}"

        while True:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for wallet in response.data:
                array_of_wallets.append(WalletInstance(self.auth_instance, **wallet))
            if response.raw.json()["next"]:
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [MAX_PAGE_SIZE]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_wallets


class TransactionInstance:
    def __init__(self, **transaction_attr):
        self.path = "/kms/wallets/"
        self.id = transaction_attr.get("id")
        self.txhash = transaction_attr.get("txhash")
        self.sender = transaction_attr.get("sender")
        self.recipient = transaction_attr.get("recipient")
        self.asset_id = transaction_attr.get("asset_id")
        self.quantity = int(transaction_attr.get("quantity")) if transaction_attr.get("quantity") is not None else None
        self.fee = int(transaction_attr.get("fee")) if transaction_attr.get("fee") is not None else None
        self.status = transaction_attr.get("status")


class Transactions:
    def __init__(self, auth_instance, wallet_id):
        self.path = "/kms/wallets/"
        self.auth_instance = auth_instance
        self.wallet_id = wallet_id

    def create(self, password, asset_id, quantity, fee, recipient, asynchronous=None, fund=None):
        # Provide password and asset_id for wallet creation
        asynchronous = asynchronous if asynchronous is not None else False
        body = {
            "password": password,
            "asset_id": asset_id,
            "quantity": str(quantity),
            "fee": str(fee),
            "recipient": recipient,
            "async": asynchronous,
        }
        if fund is not None:
            body["fund"] = fund
        response = Response(
            Request().post(
                auth_instance=self.auth_instance, path=f"{self.path}{self.wallet_id}/transactions/", body=body
            )
        )
        return TransactionInstance(**response.data)

    def get(self, transaction_id):
        # Define tx endpoint
        response = Response(
            Request().get(
                auth_instance=self.auth_instance, path=f"{self.path}{self.wallet_id}/transactions/{transaction_id}"
            )
        )
        return TransactionInstance(**response.data)

    def list(self, count):
        # Retrieve subset of wallets
        MAX_PAGE_SIZE = 100
        array_of_transactions = []
        path = f"{self.path}{self.wallet_id}/transactions/?page_size={MAX_PAGE_SIZE}"
        listed_count = 0

        while listed_count < count:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for transaction in response.data:
                array_of_transactions.append(TransactionInstance(**transaction))
                listed_count += 1
                if listed_count >= count:
                    break
            if response.raw.json()["next"]:
                page_size = min(MAX_PAGE_SIZE, count - listed_count)
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [page_size]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_transactions

    def all(self):
        MAX_PAGE_SIZE = 100
        array_of_transactions = []
        path = f"{self.path}{self.wallet_id}/transactions/?page_size={MAX_PAGE_SIZE}"

        while True:
            response = Response(Request().get(auth_instance=self.auth_instance, path=path))
            for transaction in response.data:
                array_of_transactions.append(TransactionInstance(**transaction))
            if response.raw.json()["next"]:
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [MAX_PAGE_SIZE]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_transactions


class WebhookInstance:
    def __init__(self, **webhook_attr):
        self.id = webhook_attr["id"]
        self.url = webhook_attr["url"]
        self.name = webhook_attr["name"]
        self.hmac_secret_key = webhook_attr["hmac_secret_key"]
        self.headers = webhook_attr["headers"]
        self.version = webhook_attr["version"]
        self.status = webhook_attr["status"]
        self.event_filters = webhook_attr["event_filters"]


class Webhooks:
    def __init__(self, auth_instance):
        self.auth = auth_instance
        self.path = "/tenancy/webhooks/"
        self.path_verify = "/tenancy/webhooks-verify/"

    def verify(self, url):
        body = {"verify_url": url}
        resp = Response(Request().post(auth_instance=self.auth, path=self.path_verify, body=body))
        return resp.status_code == 201

    def create(self, url, name, hmac_secret_key, version, status, event_filters, headers=None):
        body = {
            "url": url,
            "name": name,
            "hmac_secret_key": hmac_secret_key,
            "headers": headers or {},
            "version": version,
            "status": status,
            "event_filters": event_filters,
        }
        response = Response(Request().post(auth_instance=self.auth, path=self.path, body=body))
        return WebhookInstance(**response.data)

    def get(self, webhook_id):
        response = Response(Request().get(auth_instance=self.auth, path=f"{self.path}{webhook_id}"))
        return WebhookInstance(**response.data)

    def delete(self, webhook_id):
        response = Response(Request().delete(auth_instance=self.auth, path=f"{self.path}{webhook_id}"))
        return response.status_code == http.HTTPStatus.NO_CONTENT

    def list(self, count):
        # Retrieve subset of webhooks
        MAX_PAGE_SIZE = 100
        array_of_webhooks = []
        path = f"{self.path}?page_size={MAX_PAGE_SIZE}"
        listed_count = 0

        while listed_count < count:
            response = Response(Request().get(auth_instance=self.auth, path=path))
            for webhook in response.data:
                array_of_webhooks.append(WebhookInstance(**webhook))
                listed_count += 1
                if listed_count >= count:
                    break
            if response.raw.json()["next"]:
                page_size = min(MAX_PAGE_SIZE, count - listed_count)
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [page_size]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_webhooks

    def all(self):
        # Retrieve subset of webhooks
        MAX_PAGE_SIZE = 100
        array_of_webhooks = []
        path = f"{self.path}?page_size={MAX_PAGE_SIZE}"

        while True:
            response = Response(Request().get(auth_instance=self.auth, path=path))
            for webhook in response.data:
                array_of_webhooks.append(WebhookInstance(**webhook))
            if response.raw.json()["next"]:
                path = response.raw.json()["next"].split(API_VERSION)[-1]
                path_parts = list(urlsplit(path))
                query = parse_qs(path_parts[3])
                query["page_size"] = [MAX_PAGE_SIZE]
                path_parts[3] = urlencode(query, doseq=True)
                path = urlunsplit(path_parts)
            else:
                break
        return array_of_webhooks


class HistoricalData:
    def __init__(self, auth_instance):
        self.auth = auth_instance
        self.path = "/data"

    def get_transactions(self, protocol, network, address, filters=None):
        """List transactions that have been sent to and received by an address"""
        filters = filters if filters else {}
        response = Response(
            Request().get(
                auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/transactions/{address}", params=filters
            )
        )
        data = response.data
        txs = []
        for tx in response.data["result"]:
            tx["sender"] = tx.pop("from")
            txs.append(HDTransaction(**tx))

        return HDTransactions(result=txs, next_cursor=data["next_cursor"])

    def get_transaction(self, protocol, network, txhash):
        """Retrieve transaction (single) by txhash"""
        response = Response(
            Request().get(auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/transaction/{txhash}")
        )
        tx = response.data["result"]
        tx["sender"] = tx.pop("from")
        return HDTransaction(**tx)

    def get_block(self, protocol, network, block_number):
        """Retrieve block details by blockNumber"""
        response = Response(
            Request().get(auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/block/{block_number}")
        )
        return HDBlock(**response.data["result"])

    def get_asset_balance(self, protocol, network, address):
        """Retrieve native asset balance by address"""
        response = Response(
            Request().get(auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/balance/{address}")
        )
        return HDBalance(**dict(**response.data["result"], contract=None))

    def get_contract_balance(self, protocol, network, address, contract_address):
        """Retrieve contract asset balance by address"""
        response = Response(
            Request().get(
                auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/balance/{address}/{contract_address}"
            )
        )
        return HDBalance(**response.data["result"])

    def get_api_status(self, protocol, network):
        """Returns the historical data API status"""
        response = Response(Request().get(auth_instance=self.auth, path=f"{self.path}/{protocol}/{network}/status"))
        return HDStatus(**response.data["result"])



import os

API_VERSION = "1.0"
OAUTH_PATH = "/clientele/oauth2/token"
ENCODING = "utf-8"
GRANT_TYPE = "password"
SCOPE = "read write echo transaction"

UPVEST_API_TARGET = os.getenv("UPVEST_API_TARGET", default="https://api.playground.upvest.co/")
PRIVATE_KEY_BASE64 = os.getenv("PRIVATE_KEY_BASE64")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: recoverykit.proto

# flake8: noqa

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="recoverykit.proto",
    package="co.upvest.recoverykit",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\x11recoverykit.proto\x12\x15\x63o.upvest.recoverykit"E\n\x0c\x43ipherParams\x12\x13\n\x0b\x65phemeralpk\x18\x01 \x01(\x0c\x12\r\n\x05nonce\x18\x02 \x01(\x0c\x12\x11\n\trecipient\x18\x03 \x01(\x0c"S\n\nHashParams\x12\x0b\n\x03len\x18\x01 \x01(\r\x12\t\n\x01m\x18\x02 \x01(\r\x12\t\n\x01p\x18\x03 \x01(\r\x12\x0c\n\x04salt\x18\x04 \x01(\x0c\x12\t\n\x01t\x18\x05 \x01(\r\x12\t\n\x01v\x18\x06 \x01(\r"\xc0\x01\n\x04Seed\x12\x0e\n\x06\x63ipher\x18\x01 \x01(\t\x12\x39\n\x0c\x63ipherparams\x18\x02 \x01(\x0b\x32#.co.upvest.recoverykit.CipherParams\x12\x12\n\nciphertext\x18\x03 \x01(\x0c\x12\x0c\n\x04hash\x18\x04 \x01(\x0c\x12\x14\n\x0chashfunction\x18\x05 \x01(\t\x12\x35\n\nhashparams\x18\x06 \x01(\x0b\x32!.co.upvest.recoverykit.HashParams"\xb7\x01\n\x0bRecoveryKit\x12)\n\x04seed\x18\x01 \x01(\x0b\x32\x1b.co.upvest.recoverykit.Seed\x12\x10\n\x08seedhash\x18\x02 \x01(\t\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x04 \x01(\x03\x12\x11\n\tclient_ip\x18\x05 \x01(\t\x12\x0f\n\x07version\x18\x06 \x01(\t\x12\x12\n\nuser_agent\x18\x07 \x01(\t\x12\x0f\n\x07user_id\x18\x08 \x01(\rb\x06proto3'
    ),
)


_CIPHERPARAMS = _descriptor.Descriptor(
    name="CipherParams",
    full_name="co.upvest.recoverykit.CipherParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="ephemeralpk",
            full_name="co.upvest.recoverykit.CipherParams.ephemeralpk",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="nonce",
            full_name="co.upvest.recoverykit.CipherParams.nonce",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="recipient",
            full_name="co.upvest.recoverykit.CipherParams.recipient",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=44,
    serialized_end=113,
)


_HASHPARAMS = _descriptor.Descriptor(
    name="HashParams",
    full_name="co.upvest.recoverykit.HashParams",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="len",
            full_name="co.upvest.recoverykit.HashParams.len",
            index=0,
            number=1,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="m",
            full_name="co.upvest.recoverykit.HashParams.m",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="p",
            full_name="co.upvest.recoverykit.HashParams.p",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="salt",
            full_name="co.upvest.recoverykit.HashParams.salt",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="t",
            full_name="co.upvest.recoverykit.HashParams.t",
            index=4,
            number=5,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="v",
            full_name="co.upvest.recoverykit.HashParams.v",
            index=5,
            number=6,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=115,
    serialized_end=198,
)


_SEED = _descriptor.Descriptor(
    name="Seed",
    full_name="co.upvest.recoverykit.Seed",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="cipher",
            full_name="co.upvest.recoverykit.Seed.cipher",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="cipherparams",
            full_name="co.upvest.recoverykit.Seed.cipherparams",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="ciphertext",
            full_name="co.upvest.recoverykit.Seed.ciphertext",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="hash",
            full_name="co.upvest.recoverykit.Seed.hash",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="hashfunction",
            full_name="co.upvest.recoverykit.Seed.hashfunction",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="hashparams",
            full_name="co.upvest.recoverykit.Seed.hashparams",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=201,
    serialized_end=393,
)


_RECOVERYKIT = _descriptor.Descriptor(
    name="RecoveryKit",
    full_name="co.upvest.recoverykit.RecoveryKit",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="seed",
            full_name="co.upvest.recoverykit.RecoveryKit.seed",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="seedhash",
            full_name="co.upvest.recoverykit.RecoveryKit.seedhash",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="username",
            full_name="co.upvest.recoverykit.RecoveryKit.username",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="datetime",
            full_name="co.upvest.recoverykit.RecoveryKit.datetime",
            index=3,
            number=4,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="client_ip",
            full_name="co.upvest.recoverykit.RecoveryKit.client_ip",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="version",
            full_name="co.upvest.recoverykit.RecoveryKit.version",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="user_agent",
            full_name="co.upvest.recoverykit.RecoveryKit.user_agent",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="user_id",
            full_name="co.upvest.recoverykit.RecoveryKit.user_id",
            index=7,
            number=8,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=396,
    serialized_end=579,
)

_SEED.fields_by_name["cipherparams"].message_type = _CIPHERPARAMS
_SEED.fields_by_name["hashparams"].message_type = _HASHPARAMS
_RECOVERYKIT.fields_by_name["seed"].message_type = _SEED
DESCRIPTOR.message_types_by_name["CipherParams"] = _CIPHERPARAMS
DESCRIPTOR.message_types_by_name["HashParams"] = _HASHPARAMS
DESCRIPTOR.message_types_by_name["Seed"] = _SEED
DESCRIPTOR.message_types_by_name["RecoveryKit"] = _RECOVERYKIT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CipherParams = _reflection.GeneratedProtocolMessageType(
    "CipherParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _CIPHERPARAMS,
        "__module__": "recoverykit_pb2"
        # @@protoc_insertion_point(class_scope:co.upvest.recoverykit.CipherParams)
    },
)
_sym_db.RegisterMessage(CipherParams)

HashParams = _reflection.GeneratedProtocolMessageType(
    "HashParams",
    (_message.Message,),
    {
        "DESCRIPTOR": _HASHPARAMS,
        "__module__": "recoverykit_pb2"
        # @@protoc_insertion_point(class_scope:co.upvest.recoverykit.HashParams)
    },
)
_sym_db.RegisterMessage(HashParams)

Seed = _reflection.GeneratedProtocolMessageType(
    "Seed",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEED,
        "__module__": "recoverykit_pb2"
        # @@protoc_insertion_point(class_scope:co.upvest.recoverykit.Seed)
    },
)
_sym_db.RegisterMessage(Seed)

RecoveryKit = _reflection.GeneratedProtocolMessageType(
    "RecoveryKit",
    (_message.Message,),
    {
        "DESCRIPTOR": _RECOVERYKIT,
        "__module__": "recoverykit_pb2"
        # @@protoc_insertion_point(class_scope:co.upvest.recoverykit.RecoveryKit)
    },
)
_sym_db.RegisterMessage(RecoveryKit)


# @@protoc_insertion_point(module_scope)



import os

import setuptools

with open("upvest/__pkginfo__.py") as f:
    exec(f.read())
_VERSION = globals()["__version__"]


_README = open(os.path.join(os.path.dirname(__file__), "README.md")).read()

_INSTALL_REQUIRES = ("requests>=2.21.0,<3", "environs==4.1.0")
_OPTIONAL_REQUIRES = {
    "dev": ("pre-commit==1.10.5", "prospector==1.1.6.2"),
    "test": ("pytest==5.2.1", "py-ecc==1.7.0", "pysha3==1.0.2", "bitcoin==1.1.42", "web3==5.5.0"),
    "recovery": ("protobuf>=3.9.1", "PyNaCl>=1.3.0"),
}

_CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

_KEYWORDS = ["api", "upvest", "bitcoin", "ethereum", "oauth2", "client"]

_PACKAGES = setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"])

setuptools.setup(
    name="upvest",
    version=_VERSION,
    author="Upvest GmbH",
    author_email="tech@upvest.co",
    maintainer="Alexander Reichhardt",
    maintainer_email="alexander@upvest.co",
    description="Upvest API client library",
    keywords=_KEYWORDS,
    long_description=_README,
    long_description_content_type="text/markdown",
    url="https://github.com/upvestco/upvest-python/",
    packages=_PACKAGES,
    include_package_data=True,
    classifiers=_CLASSIFIERS,
    install_requires=_INSTALL_REQUIRES,
    extras_require=_OPTIONAL_REQUIRES,
    license="MIT",
)
